/*
 * Copyright (c) 1995-2002 Conexant Systems, Inc.
 * 
 * 1.   Permitted use. Redistribution and use in source and binary forms,
 * with or without modification, are permitted under the terms set forth
 * herein.
 * 
 * 2.   Disclaimer of Warranties. CONEXANT AND OTHER CONTRIBUTORS MAKE NO
 * REPRESENTATION ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.
 * IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTIES OF ANY KIND.
 * CONEXANT AND OTHER CONTRIBUTORS DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, GOOD TITLE AND AGAINST INFRINGEMENT.
 * 
 * This software has not been formally tested, and there is no guarantee that
 * it is free of errors including, but not limited to, bugs, defects,
 * interrupted operation, or unexpected results. Any use of this software is
 * at user's own risk.
 * 
 * 3.   No Liability.
 * 
 * (a) Conexant or contributors shall not be responsible for any loss or
 * damage to Company, its customers, or any third parties for any reason
 * whatsoever, and CONEXANT OR CONTRIBUTORS SHALL NOT BE LIABLE FOR ANY
 * ACTUAL, DIRECT, INDIRECT, SPECIAL, PUNITIVE, INCIDENTAL, OR CONSEQUENTIAL
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED, WHETHER IN CONTRACT, STRICT OR OTHER LEGAL THEORY OF
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 * 
 * (b) User agrees to hold Conexant and contributors harmless from any
 * liability, loss, cost, damage or expense, including attorney's fees,
 * as a result of any claims which may be made by any person, including
 * but not limited to User, its agents and employees, its customers, or
 * any third parties that arise out of or result from the manufacture,
 * delivery, actual or alleged ownership, performance, use, operation
 * or possession of the software furnished hereunder, whether such claims
 * are based on negligence, breach of contract, absolute liability or any
 * other legal theory.
 * 
 * 4.   Notices. User hereby agrees not to remove, alter or destroy any
 * copyright, trademark, credits, other proprietary notices or confidential
 * legends placed upon, contained within or associated with the Software,
 * and shall include all such unaltered copyright, trademark, credits,
 * other proprietary notices or confidential legends on or in every copy of
 * the Software.
 * 
 */
#ifndef _RIPTIDEM_H_
#define _RIPTIDEM_H_

#if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
#define SUPPORT_SOUNDCORE

#define SUPPORT_WAVE_IN
#define SUPPORT_WAVE_OUT
#define SUPPORT_MIXER

#define SUPPORT_MICROPHONE
#define SUPPORT_CD_ROM
#define SUPPORT_LINE_IN
#define SUPPORT_VIDEO
#define SUPPORT_AUX
#endif

#if defined(CONFIG_SOUND_OPL3SA1) || defined(CONFIG_SOUND_OPL3SA1_MODULE) || \
		defined(CONFIG_SOUND_OPL3SA2) || defined(CONFIG_SOUND_OPL3SA2_MODULE) ||\
		defined(CONFIG_SOUND_YM3812) || defined(CONFIG_SOUND_YM3812_MODULE) ||\
		defined(CONFIG_SOUND_ADLIB) || defined(CONFIG_SOUND_ADLIB_MODULE) ||\
		defined(CONFIG_SOUND_YMFPCI_LEGACY)
#define SUPPORT_LEGACY_DEVICES
#define SUPPORT_SYNTH
#endif

#if defined(CONFIG_SOUND_MIDI) || \
		defined(CONFIG_SOUND_MPU401) || defined(CONFIG_SOUND_MPU401_MODULE)
#define SUPPORT_MIDI_IN
#define SUPPORT_MIDI_OUT
#endif

#if defined(CONFIG_JOYSTICK) || defined(CONFIG_JOYSTICK_MODULE)
#define SUPPORT_JOYSTICK
#endif

#define SUPPORT_MODEM

#if DBG

#define mdprintf(t) printk t
#define mdprintf2(t) printk t

#else

#define mdprintf(t)
#define mdprintf2(t)

#endif

typedef struct _rt_func_table
{
    void * (*RipTideNextChannel)(void *pHwInfo, void **pContext);
    BOOLEAN (*RipTideQueryInterrupt)(
				void *pSGC, system_objects *sysObj
				);
    VOID    (*RipTideHwCompleteInterrupt)(
                void *pHwInfo);
    VOID    (*RipTideDeleteChannel)( 
                void *pChannelInfo, system_objects *sysObj
                );
    void *  (*RipTideCreateChannel)(
			    void   *pHwInfo,
			    int     iCapture,
			    void   *pContext, system_objects *sysObj);
    size_t  (*RipTideAddBuffer)( 
			    void     *pChannelInfo, 
			    PVOID     pData, 
			    ULONG     ulByteCount 
			);
    BOOLEAN (*RipTideHwStartDMA)( 
				void *pChannelInfo 
			);
    BOOLEAN (*RipTideHwStopDMA)( 
				void *pChannelInfo 
			); 
    ULONG   (*RipTideHwReadDMACounter)( 
				void *pChannelInfo 
				); 
    void    (*RipTideGetFragment)(
				void *pChannelInfo,
				int  *fragsize,
				int  *fragstotal
				); 
    void    (*RipTideSetFragment)(
				void *pChannelInfo,
				int   fragsize,
				int   fragstotal
				); 
	int     (*RipTideQueryBusy)(
			    void *pChannelInfo
			);
	int     (*RipTideQueryIdle)(
			    void *pChannelInfo
			);                
    void    (*RipTideSetSampleRate)(
			    void *pChannelInfo,
			    int   SamplesRate
			);
    int     (*RipTideGetSampleRate)(
			    void *pChannelInfo
			);
    void    (*RipTideSetBitsPerSample)(
			    void *pChannelInfo,
			    int   BitsPerSample
			);
	int     (*RipTideGetBitsPerSample)(
			    void *pChannelInfo
		    );
    void    (*RipTideSetChannels)(
			    void *pChannelInfo,
			    int   Channels
			);
	int     (*RipTideGetChannels)(
		        void *pChannelInfo
			);
    BOOLEAN (*RipTideHwGetMixer)( 
				PVOID     pHwInfo, 
				WORD      wId, 
				WORD      *pValue);
	BOOLEAN	(*RipTideHwSetMixer)( 
				PVOID     pHwInfo, 
				WORD      wId, 
				WORD      *pValue 
				);
	void	(*RipTideHwRelease)
			(   void *pHwInfo
			);
    PVOID   (*RipTideHwGetDMABuffer)(
		        void *pChannelInfo,
		        int  *size
			);
    

} rt_func_table;



typedef struct _rt_state
{
	/* we keep cm cards in a linked list */
	struct _rt_state *next;

    int dev_type;       /* Type of device: 0 - RipTide, 1 - ICH */
	/* soundcore stuff */
	int dev_audio;
	int dev_mixer;
	int dev_midi;
	int dev_dmfm;

    unsigned int            hwIOBase;       /* Riptide Base I/O address*/
    unsigned int            hwIOBase2;       /* Riptide Base I/O address*/
    unsigned int            hwIOSynth;      /* Riptide Synth I/O address*/
    unsigned int            hwIOMidi;       /* Riptide MIDI I/O address*/
    unsigned int            hwIrq;          /* RipTide IRQ */

    struct tq_struct        run_bh;
    int                     modify_counter;

    rt_func_table           *func_table;
	void                    *pHwInfo;
#ifdef MONOLIT
    void                    (*clean)(void);
#else
	char                    *modulename;
#endif
    void                    (*open)(void);
    void                    (*close)(void);
    void                    *pHwContext;
#if defined(SUPPORT_MIDI_IN) || defined (SUPPORT_MIDI_IN)
    int                     midiSlot;
#endif
    struct pci_dev          *pcidev;
} rt_state;

typedef void (*interrupt_handler)(int irq, void *dev_id, struct pt_regs *regs);

#ifdef SUPPORT_MIXER
#include "rtmixer.h"
//==================================
// Multiplexor Selects
#define RWASM_MIC    0
#define RWASM_CD     1         
#define RWASM_VIDEO  2
#define RWASM_AUX    3
#define RWASM_LINE   4
#define RWASM_MIXER  5
#define RWASM_MONO   6
#define RWASM_PHN    7    
#endif

#endif

